"""Docs imports."""

from .. import core

__all__ = (
    'configparser',
    'shutil',
    'string',
    *core.lib.__all__
    )

import configparser
import shutil
import string

from .. core . lib import *
