"""
Overview
========

**Author:** dan@1howardcapital.com

**Summary:** Zero-dependency python framework for object oriented development.

---

Usage
-----

```py
import fqr

```

"""

__all__ = (
    'cfg',
    'codecs',
    'enm',
    'exc',
    'lib',
    'strings',
    'typ',
    )

from . import cfg
from . import codecs
from . import enm
from . import exc
from . import lib
from . import strings
from . import typ
