"""Constant values specific to module unit tests."""

__all__ = (
    'Constants',
    )

from fqr import loggers


class Constants(loggers.cfg.Constants):
    """Constant values specific to unit tests in this module."""
